-- Welcome to this walk-through of Snowflake array functionality! This is a series of exercises
-- intended to enhance your understanding of the Snowflake array datatype, particularly in the context
-- of the new ARRAY_DISTINCT function that was a part of Snowflake 6.36 release (Nov 2022).

-- To complete this walkthrough, we recommend copy-pasting the full contents of this document
-- into a worksheet in your Snowflake instance. You do not need to connect to any data sources,
-- as we will be generating simple sample arrays as a part of this exercise, and you do not
-- need any "creation" or engineering permissions (we will not be creating any tables, functions,
-- stages, pipes, views, etc.) You will need a running warehouse for the queries to compute;
-- an XS warehouse will be more than sufficient.

-- Go through these instructions from top to bottom and execute each query in turn.
-- Look at the functions being used and the inputs and outputs to be sure you understand
-- before moving on to the next query. You may also decide to alter the statements and test different
-- inputs, but we do suggest starting with the provided examples and building from there.

-- Some things in particular to pay attention to as you are introduced to a new function:
-- What are the inputs of the function? e.g. are you inputting an individual datapoint, an array,
-- or multiple of one or the other? Do the inputs need to be the same datatype or same size?
-- What is the output of the function? (In this walk-through, in most cases, the output is an array.)
-- How many inputs can the function take? We can see based on the first example below that ARRAY_CONSTRUCT
-- is able to accept at least three inputs (in that case, three individual values).  Later, we will see
-- that ARRAY_CAT can only accept two array inputs, and ARRAY_DISTINCT itself only accepts one array input.
-- Finally, is the function an aggregation function, operating across multiple rows of data?
-- Or is it a basic function performing on a single record or value?

-- First, create a simple array. You can use:
-- 1) The "ARRAY_CONSTRUCT" function
-- 2) Square brackets, which are Snowflake array constants

SELECT
    ARRAY_CONSTRUCT('A','B','C') AS arr_funct,
    ['A', 'B', 'C'] AS arr_const
;

-- Try to create a more complex array:
-- Note: arrays use a semi-structured datatype known as a variant, which can contain any other datatype.
-- When you use ARRAY_CONSTRUCT and similar functions, all elements of the array are automatically cast to variant.

-- 1) An array with different datatypes (before they all get auto-cast to variant)

SELECT
    ARRAY_CONSTRUCT('A', 1, true, 'D') AS arr1
;

-- 2) An array with duplicate values

SELECT
    ARRAY_CONSTRUCT('A', 'C', 'C', 'D') AS arr1
;

-- 3) An array with missing, empty, or null values

SELECT
    ARRAY_CONSTRUCT('A', 'B', 'C', 'D', null, '', ' ') AS arr1
;

-- 4) A nested array (an array made up of two or more arrays)

SELECT
    ARRAY_CONSTRUCT(
        ARRAY_CONSTRUCT('a', 1),
        ARRAY_CONSTRUCT('b', 2),
        ARRAY_CONSTRUCT('b', 2)
    ) AS arr1
;

-- Now create a dataset with one column ('val') and four rows, values A, B, C, and D

SELECT 'A' AS val UNION ALL
SELECT 'B' AS val UNION ALL
SELECT 'C' AS val UNION ALL
SELECT 'D' AS val
;

-- Create a table with two columns (ind and val) and four rows, with values 1, 2, 3, 4 corresponding to A, B, C, and D

SELECT 1 AS ind, 'A' AS val UNION ALL
SELECT 2 AS ind, 'B' AS val UNION ALL
SELECT 3 AS ind, 'C' AS val UNION ALL
SELECT 4 AS ind, 'D' AS val
;

-- Let's add some complexity: a fifth row with values 5 and A, and a sixth row with values 2 and B

SELECT 1 AS ind, 'A' AS val UNION ALL
SELECT 2 AS ind, 'B' AS val UNION ALL
SELECT 3 AS ind, 'C' AS val UNION ALL
SELECT 4 AS ind, 'D' AS val UNION ALL
SELECT 5 AS ind, 'A' AS val UNION ALL
SELECT 2 AS ind, 'B' AS val
;

-- Use the ARRAY_AGG function to aggregate the data in this generated table into:
-- an array, a distinct array, a nested array, and a distinct nested array

SELECT
  ARRAY_AGG(val) AS arr,
  ARRAY_AGG(DISTINCT val) AS arr_dist,
  ARRAY_AGG(ARRAY_CONSTRUCT(ind, val)) AS arr_nst,
  ARRAY_AGG(DISTINCT ARRAY_CONSTRUCT(ind, val)) AS arr_nst_dist
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val
) AS smpl
;

-- Add a seventh row with values 6 and null

SELECT 1 AS ind, 'A' AS val UNION ALL
SELECT 2 AS ind, 'B' AS val UNION ALL
SELECT 3 AS ind, 'C' AS val UNION ALL
SELECT 4 AS ind, 'D' AS val UNION ALL
SELECT 5 AS ind, 'A' AS val UNION ALL
SELECT 2 AS ind, 'B' AS val UNION ALL
SELECT 6 AS ind, null AS val
;

-- Use the ARRAY_AGG function to aggregate the data in this generated table into:
-- an array, a distinct array, a nested array, and a distinct nested array.

SELECT
  ARRAY_AGG(val) AS arr,
  ARRAY_AGG(DISTINCT val) AS arr_dist,
  ARRAY_AGG(ARRAY_CONSTRUCT(ind, val)) AS arr_nst,
  ARRAY_AGG(DISTINCT ARRAY_CONSTRUCT(ind, val)) AS arr_nst_dist
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 6 AS ind, null AS val
) AS smpl
;

-- You can also see that ARRAY_AGG, as the name gives away, is an aggregate function.
-- Try returning an array and a distinct array of the 'val' field, selecting and grouping by the 'ind' field:

SELECT
  ind,
  ARRAY_AGG(val) AS arr,
  ARRAY_AGG(DISTINCT val) AS arr_dist
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 6 AS ind, null AS val
) AS smpl
GROUP BY 1
;

--  And, to close out our exploration of "ARRAY_AGG", you can also order the values
-- within an array generated by "ARRAY_AGG" using "WITHIN GROUP (ORDER BY)" syntax.
-- However, the value you are ordering by must either be the same value that you are
-- aggregating (e.g. 'val'):

SELECT
  ARRAY_AGG(val) WITHIN GROUP(ORDER BY val) AS arr,
  ARRAY_AGG(DISTINCT val) WITHIN GROUP(ORDER BY val) AS arr_dist
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 6 AS ind, null AS val
) AS smpl
;

-- ... or you must omit the DISTINCT:

SELECT
  ARRAY_AGG(val) WITHIN GROUP(ORDER BY ind) AS arr
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 6 AS ind, null AS val
) AS smpl
;

-- ... in which case, you would need to use a different function to remove the duplicate 'B' and 'A' values.

-- Which brings us back to ARRAY_DISTINCT!

-- Selecting two of the unnested arrays created above that have at least one value in common, combine the two arrays using ARRAY_CAT:

SELECT
    ARRAY_CAT(
        ARRAY_CONSTRUCT('A', 'B', 'C'),
        ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
    ) AS arr1
;

-- You can also combine three arrays using two layers of ARRAY_CAT:

SELECT
    ARRAY_CAT(
        ARRAY_CAT(
            ARRAY_CONSTRUCT('A', 'B', 'C'),
            ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
        ),
        ARRAY_CONSTRUCT(null, 1, 'DAS42')
    ) AS arr1
;

-- Now, given two arrays -- one containing A, B, and C, the other containing A, C, C, D, and undefined (null) --
-- say you would like to return a single array, with values A, B, C, D, and undefined (null).

-- In the past, before Snowflake release 6.36, what you would most likely have done is combined the two arrays, flattened them,
-- and then re-aggregated them distinctly using ARRAY_AGG(DISTINCT) as shown above. However, array_agg(distinct) ignores null values,
-- so it would only return A, B, C, and D, like so:

SELECT
    ARRAY_AGG(DISTINCT arr."VALUE") AS arr_dist
FROM TABLE(FLATTEN(
  ARRAY_CAT(
        ARRAY_CONSTRUCT('A', 'B', 'C'),
        ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
  )
)) AS arr;

-- You might have chosen to manually re-append an undefined / null value, like so:

SELECT
    ARRAY_APPEND(
      ARRAY_AGG(DISTINCT arr."VALUE"),
    null) AS arr_dist
FROM TABLE(FLATTEN(
  ARRAY_CAT(
        ARRAY_CONSTRUCT('A', 'B', 'C'),
        ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
  )
)) AS arr;

-- We're sure clever developers with a use case that required this may have been able to find other ways
-- to get the desired result, but regardless, it was certainly not ideal!

-- Instead of flattening and reaggregating, we can now use ARRAY_DISTINCT, which takes as its argument an array and returns it
-- with distinct values, including null:

SELECT
    ARRAY_DISTINCT(
        ARRAY_CAT(
            ARRAY_CONSTRUCT('A', 'B', 'C'),
            ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
        )
    ) AS with_distinct
;

-- And if you do NOT want the null values, you can merely use ARRAY_DISTINCT along with the function "ARRAY_COMPACT":


SELECT
    ARRAY_COMPACT(
        ARRAY_DISTINCT(
            ARRAY_CAT(
                ARRAY_CONSTRUCT('A', 'B', 'C'),
                ARRAY_CONSTRUCT('A', 'C', 'C', 'D', null)
            )
        )
    ) AS with_distinct_compact
;

-- Finally, as we called out on the main blog post about the ARRAY_DISTINCT release, be mindful that ARRAY_DISTINCT
-- will only deduplicate identical records. In the case of nested records, take a slight modification to the previous example:

SELECT
  ARRAY_AGG(
    ARRAY_CONSTRUCT(ind, val)
  ) AS arr_nst
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 5 AS ind, null AS val
) AS smpl
;

-- The value [2, 'A'] is duplicated here, but the value [5, 'A'] is not.  Despite the fact that there is
-- another record with ind 5 and another record with val A, that combination of values is unique, therefore
-- ARRAY_DISTINCT will result in removing the duplicate [2, 'B'] but not the [5, undefined].

SELECT
  ARRAY_DISTINCT(
      ARRAY_AGG(
          ARRAY_CONSTRUCT(ind, val)
      )
  ) AS arr_nst_dist
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 5 AS ind, null AS val
) AS smpl
;

-- You may notice that this example produced the same result as the previous example of performing ARRAY_AGG(DISTINCT)
-- on a nested array. In this example, you can see that BOTH of these functions will preserve a nested array with all
-- null/undefined values. A null on its own will be excluded from ARRAY_AGG() and not from ARRAY_DISTINCT. In a nested array,
-- The element that is being "distincted" is *an array* (which contains nulls), not a null itself.

SELECT
  ARRAY_DISTINCT(
      ARRAY_AGG(
          ARRAY_CONSTRUCT(ind, val)
      )
  ) AS arr_nst_dist,
  ARRAY_AGG(DISTINCT
      ARRAY_CONSTRUCT(ind, val)
  ) AS arr_nst_dist2
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 5 AS ind, null AS val UNION ALL
    SELECT null AS ind, null AS val
) AS smpl
;

-- A further demonstration of the null handling can be seen in this query, which shows the difference between
-- a) the output of "ARRAY_AGG(DISTINCT)" performed directly on 'val' (so that each element of the array is a variant
-- containing one value), and
-- b) the output of an array_agg(distinct) performed on *an array containing val* (so that each element of the array
-- is a variant *containing an array*... whose element, in turn, is a variant containing one value).

SELECT
    ARRAY_AGG(DISTINCT
              val
          ) AS arr_dist_val,
    ARRAY_AGG(DISTINCT
              ARRAY_CONSTRUCT(val)
          ) AS arr_dist_array
FROM (
    SELECT 1 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 3 AS ind, 'C' AS val UNION ALL
    SELECT 4 AS ind, 'D' AS val UNION ALL
    SELECT 5 AS ind, 'A' AS val UNION ALL
    SELECT 2 AS ind, 'B' AS val UNION ALL
    SELECT 5 AS ind, null AS val UNION ALL
) AS smpl
;

-- It is worth noting that when you are performing these functions on a real dataset, you are most likely already
-- beginning with array objects, rather than having to construct or aggregate them! Therefore, there is likely not as much need
-- for using multi-level nested functions as we have in this walk-through.


-- We hope you've enjoyed this walk-through of some Snowflake array functions, intended to demonstrate the value and usage
-- of the ARRAY_DISTINCT function released as a part of Snowflake 6.36. If you'd like to see more of this kind of content,
-- or if you need the help of Snowflake professionals, drop us a line at hello@das42.com.
